@@ -52,9 +52,9 @@ def kubectl_dump(command: PathLike | Executable, output_file: PathLike, mode: st
     error_file = f'{output_file}.err'
     subprocess_stream(command, output_file=output_file, error_file=error_file)
 
-    logger.warning(Path(error_file).read_text())
-
-    if os.path.getsize(error_file) == 0:
+    if err_text := Path(error_file).read_text().strip():
+        logger.warning(err_text)
+    else:
         os.remove(error_file)
 
 
@@ -82,22 +82,16 @@ def _parse_remote_path(_str: PathLike) -> tuple[str, str]:
 def kubectl_cp(
         src: PathLike,
         dest: PathLike,
-        err_file: Optional[PathLike] = None,
         mode: Optional[str] = 'dry-run',
         show_cmd=True,
         logger: Optional[logging.Logger] = None,
         retries: Optional[int] = 3,
 ) -> None:
-    """cp command on top of `kubectl cp` and `kubectl exec`, due to BKP-30. Departures from `kubectl cp`:
-
-    1. Portability for single file transfers: `kubectl cp` uses `tar` under the hood for any src/dest.
-       For single file transfers, we stream directly to the target, thus dropping the need for `tar`.
-    2. Functionality: `kubectl cp` does not support remote-to-remote transfers. We do, downloading to local host first.
-       Local host must have enough disk to store x2 the content size
+    """wrapper on top of `kubectl cp` with more error handling and remote-to-remote transfer functionality. Host must
+    have enough disk to store x2 the content size, as we use local fs to fully store contents before uploading to remote
 
     :param src: source path. Can be local or remote
     :param dest: target path. Can be local or remote
-    :param err_file: optional error file to write the error stream
     :param mode: execution mode. 'dry-run' only shows command, any other mode executes
     :param show_cmd: if True, logs the commands that are executed
     :param logger: logger object
@@ -125,28 +119,28 @@ def kubectl_cp(
             temp_file = dl_dest = up_src = tempfile.NamedTemporaryFile(delete=False).name
             logger.warning('Both source and target are remote. Downloading to local host first, and then uploading')
         if is_download:
-            _dl(dl_src, dl_dest, str(err_file or f'{dl_dest}.err'), mode=mode, show_=show_cmd, logger=logger, retries=retries)
+            _dl(dl_src, dl_dest, mode=mode, show_=show_cmd, logger=logger, retries=retries)
         if is_upload:
-            _up(up_src, up_dest, str(err_file or f'{up_src}.err'), mode=mode, show_=show_cmd, logger=logger, retries=retries)
+            _up(up_src, up_dest, mode=mode, show_=show_cmd, logger=logger, retries=retries)
     finally:
         if temp_file and os.path.exists(temp_file):
             os.remove(temp_file)
 
-def _dl(src: str, dest: str, err_file: str, mode: str, show_: bool, logger: logging.Logger, retries: int) -> None:
+def _dl(src: str, dest: str, mode: str, show_: bool, logger: logging.Logger, retries: int) -> None:
     remote, s_path = _parse_remote_path(src)
 
-    kubectl = Executable('kubectl', 'kubectl exec', remote, '--')
-    source_path_size = kubectl.run('du -sh', s_path, check=False)
-    if source_path_size.returncode:
-        logger.error(source_path_size.stderr)
-        raise RuntimeError(source_path_size.stderr)
+    kubectl = Executable('kubectl')
+    resp = kubectl.run('exec', remote, '-- du -sh', s_path, check=False)
+    if resp.returncode:
+        logger.error(resp.stderr)
+        raise RuntimeError(resp.stderr)
 
     import shutil
 
     from .units import Unit
 
     dest_dir = os.path.dirname(dest)
-    needed_disk =  Unit(source_path_size.stdout.split()[0])
+    needed_disk =  Unit(resp.stdout.split()[0])
     available_disk = Unit(f'{shutil.disk_usage(dest_dir).free} B')
 
     if needed_disk > available_disk:
@@ -154,90 +148,18 @@ def _dl(src: str, dest: str, err_file: str, mode: str, show_: bool, logger: logg
         logger.error(msg)
         raise OSError(msg)
 
-    if kubectl.run('test -d', s_path, check=False).returncode == 0:
-        Executable('kubectl').stream(f'cp --{retries=}', src, dest, mode=mode, show_cmd=show_)
-        return
-
-    if kubectl.run('test -f', s_path, check=False).returncode != 0:
-        logger.warning(f'Source path: {s_path}')
-        s_type = 'not a dir, file or pipe' if kubectl.run('test -p', s_path, check=False).returncode else 'a pipe'
-        logger.warning(f'Source path is {s_type}. Treating as file')
-
-    while retries > 0:
-        try:
-            kubectl_dump(kubectl.with_('cat', s_path), dest, mode=mode)
-            if mode == 'dry-run':
-                return
-            remote_checksum = kubectl.run('md5sum', s_path, show_cmd=False).stdout.split()[0]
-            local_checksum = md5sum(dest)
-            if local_checksum != remote_checksum:
-                msg = f' {local_checksum=}\n{remote_checksum=}\nChecksum mismatch!'
-                logger.error(msg)
-                raise RuntimeError(msg)
-            break
-        except Exception: # noqa
-            msg = f'Failed to download {src} to {dest}!'
-            if os.path.exists(err_file):
-                msg = Path(err_file).read_text()
-                logger.error(msg)
-                os.remove(err_file)
-            if os.path.exists(dest):
-                os.remove(dest)
-            if retries == 0:
-                raise RuntimeError(msg)
-            retries -= 1
-            time.sleep(10)
+    kubectl.stream(f'cp --{retries=}', src, dest, mode=mode, show_cmd=show_)
+    return
 
 
-def _up(src: str, dest: str, err_file: str, mode: str, show_: bool, logger: logging.Logger, retries: int) -> None:
-    remote, t_path = _parse_remote_path(dest)
-
-    exec_ = Executable('exec_', 'kubectl exec', remote)
-
+def _up(src: str, dest: str, mode: str, show_: bool, logger: logging.Logger, retries: int) -> None:
     source_exists = os.path.exists(src)
     if not source_exists and mode != 'dry-run':
         logger.error(f'FileNotFound: {src}')
         raise FileNotFoundError(f'{src} does not exist')
 
-    if source_exists and os.path.getsize(src) == 0:
-        logger.warning(f'{src} is empty! Skipping upload')
-        return
-
-    if source_exists and os.path.isdir(src):
-        Executable('kubectl').stream(f'cp --{retries=}', src, dest, mode=mode, show_cmd=show_)
-        return
-
-    cat_part = exec_.with_("-i -- sh -c 'cat > {}'")
-    if show_:
-        logger.info(cat_part.format(t_path) + ' < ' + src)
-
-    if mode == 'dry-run':
-        return
-    while retries > 0:
-        try:
-            with Path(src).open('rb') as f_in:
-                subprocess_stream(cat_part.format(t_path), stdin=f_in, error_file=err_file)
-            local_checksum = md5sum(src)
-            remote_checksum = exec_.run('md5sum', t_path).stdout.split()[0]
-            if local_checksum != remote_checksum:
-                msg = f' {local_checksum=}\n{remote_checksum=}\nChecksum mismatch!'
-                logger.error(msg)
-                raise RuntimeError(msg)
-            if os.path.exists(err_file):
-                logger.warning(Path(err_file).read_text())
-                if os.path.getsize(err_file) == 0:
-                    os.remove(err_file)
-            break
-        except Exception: # noqa
-            if msg := Path(err_file).read_text():
-                logger.error(msg)
-            os.remove(err_file)
-            if retries <= 0:
-                exec_.run('rm', t_path, check=False)
-                raise RuntimeError(msg)
-            retries -= 1
-            time.sleep(10)
-
+    Executable('kubectl').stream(f'cp --{retries=}', src, dest, mode=mode, show_cmd=show_)
+    return
 
 
 def print_namespace_events(namespace: str) -> None:
@@ -329,7 +251,8 @@ def get_data_from_secret(name: str, key: Optional[str] = None, namespace: Option
     cmd = 'get secret ' + name + ' --output jsonpath={.data' + f'.{key}' if key else '' + '}'
     return kubectl.run(cmd).stdout
 
-def get_namespace_from_file(file: str | Path, logger: logging.Logger, mode: str) -> str:
+
+def _get_namespace_from_file(file: str | Path, logger: logging.Logger, mode: str) -> str:
     """Get namespace from file, create if not present in k8s
     :param file: file to get namespace from
     :param logger: logger object
@@ -338,28 +261,32 @@ def get_namespace_from_file(file: str | Path, logger: logging.Logger, mode: str)
     """
     file_path = Path(file)
     namespace_from_file = file_path.stem.split('_')[-1]
-    if os.path.getsize(str(file)) == 0:
-        logger.warning(f'File {file} is empty')
+    file_content = file_path.read_text().strip()
+    if not file_content:
+        logger.warning(f'File {file} is empty! Using value from filename..')
         return namespace_from_file
-    namespace = ''
     try:
-        with file_path.open('r') as f:
-            content = json.load(f)
-            try:
-                namespace = content['items'][0]['metadata']['namespace']
-            except KeyError:  # single item
-                namespace = content['metadata']['namespace']
-            except IndexError:  # empty list
-                return namespace_from_file
+        content_j = json.loads(file_content)
     except json.decoder.JSONDecodeError as e:
-        logger.warning(f'JSONDecodeError: {e}. This may happen if the file is not json')
-    namespace = namespace or namespace_from_file
-    if namespace == file_path.stem:
-        logger.warning(f'Inferred namespace equals filename ({namespace}), which is suspect')
-        if mode == 'normal':
-            from . import confirm
-            confirm.default('Namespace will be created if not present in k8s. Continue?')
-    return namespace
+        logger.warning(f'JSONDecodeError: {e}. This may happen if the file is not json. Using value from filename..')
+        namespace = namespace_from_file
+        if namespace == file_path.stem:
+            logger.warning(f'Inferred namespace equals filename ({namespace}), which is suspect')
+            if mode == 'normal':
+                from . import confirm
+                confirm.default('Namespace will be created if not present in k8s. Continue?')
+        return namespace
+    try:
+        return content_j['items'][0]['metadata']['namespace']
+    except KeyError:  # single item
+        try:
+            return content_j['metadata']['namespace']
+        except KeyError:  # empty item
+            return namespace_from_file
+    except IndexError:  # empty list of k8s items
+        return namespace_from_file
+    except TypeError: # empty json list (TypeError: list indices must be integers or slices, not str)
+        return namespace_from_file
 
 
 def ensure_namespace(mode: str, logger: logging.Logger, *, namespace: Optional[str] = None,
@@ -373,13 +300,26 @@ def ensure_namespace(mode: str, logger: logging.Logger, *, namespace: Optional[s
     """
     kubectl = Executable('kubectl')
     if file:
-        namespace = get_namespace_from_file(file, logger, mode)
-    namespace_exists = kubectl.run('get namespace', namespace, check=False)
+        namespace = _get_namespace_from_file(file, logger, mode)
+    namespace_exists = kubectl.run('get namespace', namespace, '-ojsonpath={.status.phase}', check=False)
     if namespace_exists.returncode == 0:  # success
+        status = namespace_exists.stdout.strip()
+        if status not in  ['Active', 'Terminating']:
+            logger.warning(f'{namespace=}, {status=}, expected either "Active", "Terminating" or missing!')
+            raise RuntimeError(f'{namespace=}, {status=}, expected either "Active", "Terminating" or missing!')
+        if status == 'Terminating':
+            logger.warning(f'{namespace=}, {status=}. Awaiting termination and recreating it')
+            kubectl.stream('wait namespace', namespace, '--for=delete --timeout=600s')
+            return ensure_namespace(mode, logger, namespace=namespace, file=file)
         return namespace
     if namespace_exists.stderr.startswith(RESOURCE_NOT_FOUND):
         logger.warning(f'{namespace=} not found, creating...')
-        kubectl.stream('create namespace', namespace, ' --dry-run=client' if mode == 'dry-run' else '')
+        resp = kubectl.run('create namespace', namespace, ' --dry-run=client' if mode == 'dry-run' else '', check=False)
+        if 'AlreadyExists' in resp.stderr:
+            return ensure_namespace(mode, logger, namespace=namespace, file=file)
+        elif resp.returncode:
+            logger.warning(resp.stdout)
+            raise RuntimeError(f'{namespace=} failed to create!\n{resp.stderr}')
         return namespace
     if not namespace_exists.stderr.startswith('Error from server (Forbidden)'):
         raise RuntimeError(namespace_exists.stderr)
@@ -404,6 +344,7 @@ def ensure_pvc(spec: dict, logger: logging.Logger) -> None:
         logger.info('Creating persistent volume claim')
         from .templates import persistent_volume_claim
         _, path = persistent_volume_claim.generate_template(spec)
+        ensure_namespace(spec['MODE'], logger, namespace=spec['NAMESPACE'])
         kubectl.stream('create --filename', path)
 
     pvc_desired_states = [x.lower() for x in spec.get('PERSISTENT_VOLUME_CLAIM_DESIRED_STATES') or ['Bound']]
@@ -482,7 +423,7 @@ def create_oneliner_job(
     1. Ensure namespace
     2. Ensure PVC (including creating PV if needed)
     3. Create job
-    4. Redact saved job manifest yaml
+    4. Redact saved job manifest YAML
     5. Await job completion (can be separated out to Task validate phase)
 
     :param spec: dict with job parameters
@@ -547,7 +488,7 @@ def await_k8s_job_completion(spec: dict, tail: Optional[int] = None) -> bool:
     logger_plain = log.get_logger('plain')
     job_status_cmd = 'get job --output jsonpath={.status} ' + name
     output_wide_on_debug = '--output wide' if spec.get('LOG_LEVEL') == 'DEBUG' else ''
-    get_pods_cmd = f'get pods --selector=job-name={name} {output_wide_on_debug}'
+    get_pods_cmd = f'get pods --ignore-not-found --selector=job-name={name} {output_wide_on_debug}'
 
     kubectl = Executable('kubectl', f'kubectl --namespace {namespace}')
     kubectl_run = functools.partial(kubectl.run, show_cmd=False, check=False)
@@ -566,7 +507,6 @@ def await_k8s_job_completion(spec: dict, tail: Optional[int] = None) -> bool:
         response = kubectl_run(job_status_cmd)
 
     kubectl.stream(get_pods_cmd)
-    status = json.loads(kubectl.run(job_status_cmd).stdout)
 
     import itertools
     from datetime import datetime
@@ -578,14 +518,16 @@ def await_k8s_job_completion(spec: dict, tail: Optional[int] = None) -> bool:
     wait_job_cmd = f'wait job {name} --timeout={int(wait_interval)//2}s --for '
 
     response = kubectl_run(wait_job_cmd, next(conditions))
+    kubectl.stream(get_pods_cmd, '--no-headers', show_cmd=False)
     while response.returncode:
         now = datetime.now()
         if now.minute < 1 and now.second < wait_interval % 60 + 1:  # hourly liveness
             kubectl.stream(get_pods_cmd, '--no-headers', show_cmd=False)
         if response.stderr.startswith(RESOURCE_NOT_FOUND):
-            logger.debug('Job was created, but was unexpectedly deleted!\nEvents:')
+            msg = f'{response.stderr}! Was the job deleted?'
+            logger.warning(f'{msg}\nEvents:')
             print_namespace_events(namespace)
-            raise RuntimeError(f'Job was created, but was unexpectedly deleted!')
+            raise RuntimeError(msg)
         if WAIT_TIMEOUT_ERROR not in response.stderr:
             logger_plain.warning(response.stderr)
             raise RuntimeError(response.stderr)
@@ -594,20 +536,22 @@ def await_k8s_job_completion(spec: dict, tail: Optional[int] = None) -> bool:
 
     kubectl.stream(get_pods_cmd, '--no-headers', show_cmd=False)
 
-    terminal_status = json.loads(kubectl.run(job_status_cmd).stdout)
     kubectl.stream(f'logs --ignore-errors --selector=job-name={name} --since={int(wait_interval)*2}s',
                    f'--tail={tail}' if tail else '', show_cmd=False)
+    try:
+        terminal_status = json.loads(kubectl_run(job_status_cmd).stdout)
+    except json.JSONDecodeError as e:
+        msg = f'Failed to fetch job status'
+        logger.warning(msg)
+        raise RuntimeError(msg) from e
+
     if terminal_status.get('succeeded'):
         return True
-
-    if status != terminal_status: # shifting from "failed" to "unknown" etc.
-        logger.error('Running status does not match terminal status:')
-        log.log_as('json', status, printer=logger_plain.warning)
-    logger.info('Terminal status:')
+    logger.warning('Terminal status:')
     log.log_as('json', terminal_status, printer=logger_plain.warning)
-    if terminal_status.get('failed'):
-        raise RuntimeError(f'{name=}, {terminal_status=}')
-    raise RuntimeError(f'{name=}, unexpected status - {status}')
+    if not terminal_status.get('failed'):
+        logger.warning('Terminal status is unexpectedly neither succeeded nor failed')
+    raise RuntimeError(f'{name=}, {terminal_status=}')
 
 
 # deprecated
@@ -621,7 +565,7 @@ def _check_gibby_logs_for_container_hang(job_name: str, namespace: str, kubectl:
     :param logger: logger object
     :return: updated retry count
     """
-    # kubectl logs --pod-running-timeout=5m should wait for at least one pod running, but it doesn't for some reason
+    # kubectl logs --pod-running-timeout=5m should wait for at least one pod running, but it doesn't for some reason.
     # added retry as a workaround
 
     import subprocess
@@ -697,7 +641,7 @@ def _get_running_pod_manifest(kubectl: Executable, tries: int, retries: int, msg
                               logger: Optional[logging.Logger] = None) -> dict:
     """Get running pod manifest from k8s
     :param kubectl: kubectl Executable object
-    :param tries: number of tries param
+    :param tries: number of tries
     :param retries: dynamic retry counter
     :param msg: error message to raise if failed
     :param logger: logger object
@@ -715,7 +659,7 @@ def _get_running_pod_manifest(kubectl: Executable, tries: int, retries: int, msg
         items = json.loads(result_stdout).get('items')
     except json.decoder.JSONDecodeError:
         return _get_running_pod_manifest(kubectl, tries, retries, 'Failed to decode json', logger)
-    except TypeError as e:  # "'NoneType' object is not subscriptable" error means parsed json is None
+    except TypeError as e:  # "'NoneType' object is not subscriptable" error means parsed JSON is None
         return _get_running_pod_manifest(kubectl, tries, retries, str(e), logger)
     logger_plain = log.get_logger('plain')
     if not items:
@@ -746,9 +690,9 @@ def is_remote_sharing_disk_with_host(  # TODO: create test
     marker_name = f'.{spec.get("NAME") or "default"}-check-is-remote-sharing-disk-with-host-{os.urandom(4).hex()}'
     marker = Path(local_path, marker_name)
     marker.touch()
-    # -1 for single column, -A for all files except ./..
+    # -1 for single column, -A for all files except . / ..
     ls_job_name = create_oneliner_job(spec, f'ls -1A {remote_path or local_path}', 'ls', await_completion=True)
-    kubectl = Executable('kubectl', 'kubectl logs --namespace', spec['NAMESPACE'])
+    kubectl = Executable('kubectl', 'kubectl logs --tail=-1 --namespace', spec['NAMESPACE'])
     resp = kubectl.run(f'--selector=job-name={ls_job_name}', check=False)
     marker.unlink()
     return marker_name in resp.stdout.splitlines()
